uses crt, mcga, pcx;

var
    stimer : longint absolute $40:$6c;
    time: double;
    frames:integer;    
    timr,framerate : real;
    start_time,accumulated_time: longint;

const
      Zero = 0.0;
      One = 1.0;
      TwoPi = 2.0 * 3.141562;
      Pi_Over_Two = 3.141562/2.0;    

{ Returns Absolute value of x with Sign of y }
Function Sign ( x, y: double): double;
Begin
  if y >= Zero then
     Sign := Abs(x)
  else
     Sign := -Abs(x);
End;      

function arctan2(x, y: double): double;
begin
    arctan2 := (arctan(y/x));
end;

function min(a, b: double): double; 
begin
    if (a < b) then min := a else min := b;
end;

Function Atan2 (s, c: double): double;
var  s1, c1, Theta: double;
Begin
  s1 := Abs(s);
  c1 := Abs(c);
  if c1 + s1 = Zero then
    Theta := Zero
  else if s1 <= c1 then
	 Theta := ArcTan(s1 / c1)
       else
	 Theta := Pi_Over_Two - ArcTan(c1 / s1);
  if c < Zero then
    Theta := Pi - Theta;
  Atan2 := Sign(Theta, s);
End;

function sar8(a : word) : word;assembler;
asm
 mov ax, word ptr a
 sar ax, 8
end;

function sar2(a : word) : word;assembler;
asm
 mov ax, word ptr a
 sar ax, 2
end;

function sal6(a : word) : word;assembler;
asm
 mov ax, word ptr a
 sal ax, 6
end;

function dsal6(a: byte; b: byte): word; assembler; 
asm
    mov ax, 0
    mov bx, 0
    mov al, byte ptr b
    mov bl, byte ptr a
    sub ax, bx

    cmp ax, 128
     jl @@not_above

    sub ax, 256
@@not_above:

    cmp ax, -128
     jg @@not_below

    add ax, 256

@@not_below:
    sal ax, 6
end;


{ should be asm }
procedure interpolateblock(x, y: integer; tex, buff, offs: word);
var 
{px,py : integer;}
u1,v1: byte;
u2,v2: byte;
u3,v3: byte;
u4,v4: byte;
col: byte;
addr: word;
u, v: integer;
dx, dx_bottom, dy, dxdy: integer;
v_dx, v_dx_bottom, v_dy, v_dxdy: integer;
temp_u, temp_v: integer;
xx, yy: integer;
begin
    addr := (x shl 2) + (y shl 8) + (y shl 10); { x * 4 + y * 4 * 320 }

    asm 
        { TODO: Could be two 32 bit reads, instead of 4 x 16 bit }
        mov es, buff
        mov bx, y
        shl bx, 8
        mov dx, x
        shl dx, 1
        add bx, dx

        db $66; mov ax, es:[bx]
        mov u1, ah
        mov v1, al

        db $66; shr ax, 16
        mov u2, ah
        mov v2, al

        add bx, 256 {(256-2)}

        db $66; mov ax, es:[bx]
        mov u3, ah
        mov v3, al

        db $66; shr ax, 16
        mov u4, ah
        mov v4, al
    end;

    dx := dsal6(u1, u2);
    dx_bottom := dsal6(u3, u4);
    dy := dsal6(u1, u3);
    dxdy := sar2(dx_bottom - dx);
    u := u1 shl 8;
        
    v_dx := dsal6(v1, v2);
    v_dx_bottom := dsal6(v3, v4);
    v_dy := dsal6(v1, v3);
    v_dxdy := sar2(v_dx_bottom - v_dx);
    v := v1 shl 8;

    for yy := 0 to 3 do begin
        temp_u := u;
        temp_v := v;

        for xx := 0 to 3 do begin
            col := mem[tex:(temp_u shr 8)+((temp_v shr 8) shl 8)]; { (temp_u shr 8); } 
            mem[offs:addr + xx] := col; 

            temp_u := temp_u + dx;
            temp_v := temp_v + v_dx;            
        end;
        
        u := u + dy;
        v := v + v_dy;
        
        dx := dx + dxdy;
        v_dx := v_dx + v_dxdy;

        addr := addr + 320;
     end;
end;

procedure fdtun(time: double; lookup, tex, buff, offs: word);
var x,y : integer;
 col : byte;
 radius: double;
 dirx, diry, dirz, t_dirx, t_diry, t_dirz: double;
 orix, oriy, oriz: double;
 temp: double;
 a, b, c, delta: double;
 t1, t2, t: double;
 ix, iy, iz: double;
 u, v: byte;
 co, si: double;
 iu, iv: integer;
 matrix: array[0..3,0..3] of double;
 xmove: double;
 svx, svy, svz, cvx, cvy, cvz: double;
 i: byte;
 wrd: word;
begin

{
 inc(xmove,16);

    time := 1;}    
    xmove := time*0.03;

    svy:=(sin(xmove*3.1415)); cvy:=(cos(xmove*3.1415));
    svx:=(sin((xmove/2)*3.1415)); cvx:=(cos((xmove/2)*3.1415));
    svz:=(sin((xmove/3)*3.1415)); cvz:=(cos((xmove/3)*3.1415/24));

    matrix[0,0]:=cvy*cvx + svy*svz*svx;
    matrix[0,1]:=-cvy*svx + cvx*svy*svz;
    matrix[0,2]:=cvz*svy;

    matrix[1,0]:=cvz*svx;
    matrix[1,1]:=cvz*cvx;
    matrix[1,2]:=-svz;

    matrix[2,0]:=-cvx*svy + cvy*svz*svx;
    matrix[2,1]:=svy*svx + cvy*cvx*svz;
    matrix[2,2]:=cvy*cvz;


    { radius "already squared" }
    radius := 256*256;

    { ideas: 
        * precalc the 4000 normalized direction vectors (rotation is done after) (32k . 4000*8 per double)
            pack / unpack in asm?
        * interpolate "as you go", keep cache of u,v for previous line
    }

    { origin }
    orix := 1;
    oriy := 0;
    oriz := -256-(time*500);    

    c := (orix*orix) + (oriy*oriy) - (radius);

    col := 0;
    for y := 0 to 50 do begin
        for x := 0 to 80 do begin
            t_dirx := x - 40;
            t_diry := y - 25;
            t_dirz := 128;
                                    
            { matrix rotate dir }
            dirx := t_dirx*matrix[0,0] + t_diry*matrix[1,0] + t_dirz*matrix[2,0];
            diry := t_dirx*matrix[0,1] + t_diry*matrix[1,1] + t_dirz*matrix[2,1];
            dirz := t_dirx*matrix[0,2] + t_diry*matrix[1,2] + t_dirz*matrix[2,2];
            
            { solve }
            a := (dirx*dirx) + (diry*diry);
            b := 2 * ((orix*dirx) + (orix*diry));
                        
            delta := sqrt((b*b) - (4*a*c));

            t1 := (-b + delta) / (2*a);
            t2 := (-b - delta) / (2*a);

            t := min(t1, t2);

            { intersect }
            ix := orix + (t * dirx);
            iy := oriy + (t * diry);
            iz := oriz + (t * dirz);

            { tex }
            u := round(iz / 16);
            v := round(Atan2(iy, ix) * 256/TwoPi);

            asm 
                mov es, buff
                mov ah, u
                mov al, v

                mov bx, y
                shl bx, 8
                mov dx, x
                shl dx, 1
                add bx, dx
                mov es:[bx], ax
            end
        end;
     end;

    for y := 0 to 49 do begin
        for x := 0 to 79 do begin
            interpolateblock(x, y, tex, buff, offs);            
        end;
    end;

end;

procedure viewpal(offs : word);
var x,y : integer;
 col : byte;
begin
  col := 0;
  for y := 0 to 15 do begin
  for x := 0 to 15 do begin
    mem[offs:x+y*320] := col;
    inc(col);
  end;
  end; 
end;

procedure fliffpal;
var i,j: integer;
    r,g,b: byte;
begin
    { assume pcx uses only first 32 colors }
    { todo: lookup for percentage [95, 90, 80, 65, 50, 30, 10, 0]}
    for i := 1 to 7 do begin
        for j := 0 to 31 do begin
            r := palinfo[j, 1];
            g := palinfo[j, 2];
            b := palinfo[j, 3];
            palinfo[j + (i * 32), 1] := round(r * (100-(i*14)) div 100);
            palinfo[j + (i * 32), 2] := round(g * (100-(i*14)) div 100);
            palinfo[j + (i * 32), 3] := round(b * (100-(i*14)) div 100);
        end;
    end;
end;

procedure grscale;
var i: byte;
begin
  for i := 0 to 255 do begin
    palinfo[i, 1] := i div 4;
    palinfo[i, 2] := i div 4;
    palinfo[i, 3] := i div 4;
 end;
end;

procedure precalc(addr: word);
var x, y : integer;
    t_dirx, t_diry, t_dirz: double;
    temp: double;
 begin
    { iterate half }
    for y := 0 to 25 do begin
        for x := 0 to 40 do begin
            t_dirx := x - 40;
            t_diry := y - 25;
            t_dirz := 128;

            { pre totate - normalize dir }
            temp := sqrt((t_dirx*t_dirx) + (t_diry*t_diry) + (t_dirz*t_dirz));
            t_dirx := t_dirx / temp;
            t_diry := t_diry / temp;
            t_dirz := t_dirz / temp;

            asm 
                mov es, addr
                
                { + x * 24 ==> +x * 16 + x * 8}
                mov di, x
                mov bx, di
                shl di, 4
                shl bx, 5
                add di, bx

                { + y * (24 * 41) = 984 --> round up to 1024}
                mov bx, y
                shl bx, 10
                                
                add di, bx

                lea si, t_dirx

                db $66; mov ax, [si]
                db $66; mov es:[di], ax
                db $66; mov ax, [si+4]
                db $66; mov es:[di+4], ax
                            
                lea si, t_diry

                db $66; mov ax, [si]
                db $66; mov es:[di+8], ax
                db $66; mov ax, [si+4]
                db $66; mov es:[di+12], ax
                
                lea si, t_dirz

                db $66; mov ax, [si]
                db $66; mov es:[di+16], ax
                db $66; mov ax, [si+4]
                db $66; mov es:[di+20], ax
            end;
        end;
    end;
 end;

begin
    { using 4 virtual screens. babayaga uses 3, but the one here for texture could be replaced by babas texture setup }
  setupvirtual1;
  setupvirtual2;
  setupvirtual3;
  setupvirtual4;

    asm
        mov ax, 13h
        int 10h
    end;
    {

    readpcx('met3.pcx', vaddr, false);
    readpcx('text1.pcx', vaddr, false);
    readpcx('stan1.pcx', vaddr, false);
    readpcx('text.pcx', vaddr, false);
    }

    readpcx('stan_r.pcx', vaddr, false);
    fliffpal;
    {grscale;
    bandscale;}
    showpcx;
    
    time := 0;
    frames := 0;
    stimer := 0;

    Cls(vaddr2, 0);

    { precalc(vaddr2); }
    precalc(vaddr4);

    repeat
        time := time + 0.02;
        fdtun(time, vaddr4, vaddr, vaddr3, vaddr2);
        {viewpal(vaddr2);}
        WaitRetrace;
        flip(vaddr2,vga);
        inc(frames);
    until (keypressed);
  
  framerate := frames/(stimer/18.2);

  asm
    mov ax,3h
    int 10h
  end;

  writeln('FRAME RATE:',framerate:4:4);
  writeln('test', dsal6(0, 1));
end.